# -*- coding: utf-8 -*-
import logging
import re
import requests

from odoo import api, tools, models
from odoo.exceptions import UserError
from odoo.addons.saas_trial.compat import decode_message_header, extract_rfc2822_addresses
from odoo.addons.saas_worker import metabase
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.config import saas_config
from odoo.addons.saas_worker.const import MASTER, SAAS_BASE_URL
from odoo.addons.saas_worker.util import version_gte
from odoo.tools.translate import _

from .module import version_lt


_logger = logging.getLogger(__name__)
RE_EMAIL_DOM_BL = re.compile(r'example|gmail|odoo|outlook|hotmail|yahoo|aol\..*')
RE_EMAIL_DOM_WL = re.compile(r'[a-zA-Z0-9-]+\.{}'.format(SAAS_BASE_URL))

try:
    from dns.resolver import canonical_name, resolve, NXDOMAIN, NoAnswer, NoNameservers
    dns = True
except ImportError:
    _logger.info("DNS resolver not available")
    dns = False

class TemporaryMailDeliveryException(UserError):
    """Specific exception subclass for temporary mail delivery errors"""
    def __init__(self, message):
        super(TemporaryMailDeliveryException, self).__init__(message)
        self.sentry_ignored = True
        self.message = message


class IrMailServer(models.Model):
    _inherit = "ir.mail_server"

    # Show some basic info about the daily limit on the SMTP server screen
    # NOTE: only called for 16.0+
    def _compute_smtp_authentication_info(self):
        super()._compute_smtp_authentication_info()
        for server in self:
            # `smtp_authentication` field exists as of 16.0
            is_internal_smtp = (
                server.smtp_authentication == 'cli'  # as of 16.2
                or (server.smtp_host and server.smtp_host.lower().endswith('.odoo.com'))
            )
            if is_internal_smtp:
                db = metabase.get(self._cr.dbname)
                server.smtp_authentication_info = (
                    (server.smtp_authentication_info or '') + '\n\n' +
                    _('Emails sent today via Odoo servers: %s/%s') % (db.email_daily_count, db.email_daily_limit)
                )

    # override send_mail and connect depending on version (<saas-14, >saas-14)

    @api.model
    def _is_smtp_internal(self, **kwargs):
        # Determine internal or external mail server, to apply limits accordingly.
        # We assume that having no SMTP server configured, or having a server with a hostname under
        # ".odoo.com" or with auth set to 'CLI' means an internal server. The rest is external.
        # TODO: might need resolving the hostname instead of relying on the domain only

        is_internal_smtp = False

        # 1) smtp_session has precedence -> check session._host
        if kwargs.get('smtp_session'):
            # SMTP._host is a stable API
            is_internal_smtp = kwargs['smtp_session']._host.lower().endswith('.odoo.com')

        # TODO: verify whether the other steps below are still necessary. It seems that in most
        #       cases for 13.0+ the session is provided to send_email(), expect perhaps for custom
        #       server actions?

        # 2) mail_server_id -> auth 'cli' means implicit server, otherwise check host
        elif kwargs.get('mail_server_id'):
            mail_server = self.sudo().browse(kwargs['mail_server_id'])
            is_internal_smtp = (
                getattr(mail_server, 'smtp_authentication', None) == 'cli'  # possible as of 16.2
                or (mail_server.smtp_host and mail_server.smtp_host.lower().endswith('.odoo.com'))
            )

        # 3) smtp_server -> check host
        elif kwargs.get('smtp_server'):
            is_internal_smtp = kwargs['smtp_server'].lower().endswith('.odoo.com')

        # 4) fallback to first server (useful before 15.0)
        else:
            mail_server = self.sudo().search([], order='sequence', limit=1)[:1]
            if mail_server:
                is_internal_smtp = (
                    getattr(mail_server, 'smtp_authentication', None) == 'cli'  # possible as of 16.2
                    or (mail_server.smtp_host and mail_server.smtp_host.lower().endswith('.odoo.com'))
                )
            else:
                # no server = implicit internal one
                is_internal_smtp = True

        return is_internal_smtp

    @api.model
    def send_email(self, message, **kwargs):
        # each unique recipient counts as 1 mail
        recipients = []
        for f in ('To', 'Cc', 'Bcc'):
            recipients.extend(extract_rfc2822_addresses(message[f]))
        inc = len(set(recipients))

        db = metabase.get(self._cr.dbname)
        is_internal_smtp = db and self._is_smtp_internal(**kwargs)

        if is_internal_smtp and db.email_daily_count + inc > db.email_daily_limit:
            msg = ""
            if db.email_daily_limit < metabase.EMAIL_DAILY_LIMIT_PAID:
                msg = _(
                    """You have reached your daily limit of %s messages.
                    Paid subscriptions receive an increased limit of %s.
                    To raise the limit, subscribe or contact Odoo Help for a review at https://odoo.com/help"""
                ) % (db.email_daily_limit, metabase.EMAIL_DAILY_LIMIT_PAID)
            else:
                msg = _(
                    """You have reached your daily limit of %s messages.
                    To raise the limit, contact Odoo Help for a review at https://odoo.com/help"""
                ) % (db.email_daily_limit)
            _logger.warning("Daily email limit reached for db %s", self._cr.dbname)
            raise TemporaryMailDeliveryException(msg)

        res = super(IrMailServer, self).send_email(message, **kwargs)

        if is_internal_smtp:
            metabase.increment(db, email_daily_count=inc)
        else:
            metabase.increment(db, ext_email_daily_count=inc)

        return res

    def _check_odoo_domainkey(self, domain):
        """
        Verify that the odoo._domainkey CNAME is set for a custom domain. The check is considered
        valid if:
        - the check cannot be run (missing dnspython package)
        - the domain is whitelisted, i.e. @<dbname>.odoo.com
        - the CNAME is equal to odoo._domainkey.odoo.com.
        - an unexpected error occurs

        The check is invalid if:
        - the CNAME doesn't exist
        - the CNAME is not equal to odoo._domainkey.odoo.com.

        param str domain: the domain of the email address
        :return bool: True if the domainkey exists, False instead.
        """

        if not dns:
            return True
        if RE_EMAIL_DOM_WL.match(domain):
            return True
        try:
            # resolve() raises NXDOMAIN if the domain doesn't exist, while canonical_name()
            # doesn't.
            resolve("odoo._domainkey.{}".format(domain), "TXT")
            cname = canonical_name("odoo._domainkey.{}".format(domain))
        except (NXDOMAIN, NoAnswer, NoNameservers):
            _logger.info("domainkey not set for domain %s", domain)
            return False
        except Exception:
            _logger.warning("Unexpected error during domainkey resolution %s", domain, exc_info=1)
            return True
        if cname.to_text().lower() != 'odoo._domainkey.odoo.com.':
            _logger.info("domainkey incorrect for domain %s: %s", domain, cname.to_text())
            return False
        _logger.info("domainkey correct for domain %s", domain)
        return True

    def _alter_email_message_from(self, message):
        """
        Alter the mail message 'From' if it spoofs a blacklisted provider or another
        Odoo DB.

        :param `email.message` message: the message to alter
        """

        domain_from = tools.email_domain_extract(message["From"])
        # Check domainkey for information purpose. In the future, use it as a condition to alter
        # the message["From"].
        self._check_odoo_domainkey(domain_from)

        normalized_domain = "{}.{}".format(self._cr.dbname, SAAS_BASE_URL)
        replace_from = False

        if (
            RE_EMAIL_DOM_BL.match(domain_from)
            or (
                RE_EMAIL_DOM_WL.match(domain_from)
                and domain_from.split(".")[0] != self._cr.dbname
            )
        ):
            _logger.info(
                "Email From domain %s different from normalized domain %s",
                domain_from,
                normalized_domain,
            )

            # Find an existing default alias
            if "mail.alias.domain" in self.env:
                MAD = self.env["mail.alias.domain"].sudo()
                replace_from = MAD.search([("name", "=", normalized_domain)], limit=1).default_from_email

            # Fallback to ICP
            if not replace_from:
                ICP = self.env["ir.config_parameter"].sudo()
                replace_alias = ICP.get_param("mail.default.from", "notifications")
                replace_from = "{}@{}".format(replace_alias, normalized_domain)

            replace_from = tools.encapsulate_email(message["From"], replace_from)
            _logger.info("Email From changed from %s to %s", message["From"], replace_from)
            message.replace_header("From", replace_from)

    def _alter_smtp_from(self, smtp_from):
        """
        Return the smtp_from standardized with the domain `<dbname>.odoo.com`.
        If the model `mail.alias.domain` exists, we retrieve the bounce alias corresponding to
        the domain `<dbname>.odoo.com`. If the model doesn't exist, we replace the domain.

        :param str smtp_from: the smtp_from to alter
        :return str: the modified smtp_from
        """

        normalized_domain = "{}.{}".format(self._cr.dbname, SAAS_BASE_URL)
        replace_smtp_from = False

        # In 17.0+, force the value for simplicity.
        if "mail.alias.domain" in self.env:
            replace_smtp_from = "bounce@{}".format(normalized_domain)

        # Fallback to ICP
        if not replace_smtp_from:
            ICP = self.env["ir.config_parameter"].sudo()
            replace_alias = ICP.get_param("mail.bounce.alias", "bounce")
            replace_smtp_from = "{}@{}".format(replace_alias, normalized_domain)

        if replace_smtp_from != smtp_from:
            _logger.info(
                "Email smtp_from changed from %s to %s", smtp_from, replace_smtp_from
            )
            smtp_from = replace_smtp_from

        return smtp_from

    def _prepare_email_message(self, message, smtp_session):
        """
        Override intended to prevent spoofing major email providers. It alters the envelope from
        and the 'From' header.
        """

        smtp_from, smtp_to_list, message = super()._prepare_email_message(message, smtp_session)

        if self._is_smtp_internal(smtp_session=smtp_session):
            # Step 1: rewrite the From
            self._alter_email_message_from(message)

            # Step 2: rewrite the smtp_from
            smtp_from = self._alter_smtp_from(smtp_from)

        return smtp_from, smtp_to_list, message


class MailMail(models.Model):
    _inherit = "mail.mail"

    @api.multi
    def send(self, raise_exception=False, **kwargs):
        try:
            # No raise when called from the mail queue manager cron,
            # to not interrupt the mail queue loop,
            # so the iteration continues even in case of one mail being in exception.
            # `auto_commit` is `True` only when this method is called from the cron.
            should_raise = raise_exception or not kwargs.get('auto_commit')
            res = super(MailMail, self).send(raise_exception=should_raise, **kwargs)
        except TemporaryMailDeliveryException as e:
            if raise_exception or version_lt('10.0+e') or 'bus.bus' not in self.env.registry:
                # The `bus` is a dependence of mail from 9.0. Before it's a dependence of im_chat.
                # We could make the email daily limit reached warning as non-blocking from 9.0,
                # but before 10.0, that was unclear when a mail was sent or not,
                # there was no icon to inform the user the mail in a discuss thread was not sent.
                # Besides if mail is not installed (which is quite a corner case),
                # the bus is not in the registry and in this case we also raise the exception.
                raise
            else:
                author = self[0].mail_message_id.author_id
                title = _("You reached your daily limit of emails")
                message = e.message

                if version_gte('15.0'):
                    # Notifications structure was refactored in 15.0, changing the signature of _sendone
                    self.env['bus.bus']._sendone(author, 'saas_trial.notification', {
                        'title': title,
                        'message': message,
                        'sticky': True
                    })
                else:
                    # _sendone() or sendone(), depending on version
                    sendone = getattr(self.env['bus.bus'], '_sendone', None) or self.env['bus.bus'].sendone
                    sendone((self._cr.dbname, 'res.partner', author.id), {
                        'type': 'saas_trial.notification',
                        'title': title,
                        'message': message,
                        'sticky': True,
                    })
        except Exception:
            if raise_exception:
                raise
        else:
            return res


def _mailthread_handle_bounce(dbname):
    db = metabase.get(dbname)
    _logger.info("Received a bounce email for db {dbname}".format(dbname=dbname))
    metabase.increment(db, bounce_daily_count=1)


class MailThread(models.AbstractModel):
    _inherit = "mail.thread"

    @api.model
    def _detect_is_bounce(self, message, message_dict):
        """
        Override in order to detect the forced bounce address (applicable in 17.0+).
        """

        bounce_alias = "bounce@{}.{}".format(self._cr.dbname, SAAS_BASE_URL)
        email_to_list = [
            tools.email_normalize(e) or e
            for e in (tools.email_split(message_dict['to']) or [''])
        ]
        if any(email in bounce_alias for email in email_to_list):
            return True
        return super()._detect_is_bounce(message, message_dict)

    @api.model
    def message_route(self, message, message_dict, model=None, thread_id=None, custom_values=None, **kwargs):
        try:
            bounce_alias = self.env['ir.config_parameter'].sudo().get_param("mail.bounce.alias")
            email_from = decode_message_header(message, 'From')
            email_from_localpart = (tools.email_split(email_from) or [''])[0].split('@', 1)[0].lower()
            email_to = decode_message_header(message, 'To')
            email_to_localparts = [
                e.split('@', 1)[0].lower()
                for e in (tools.email_split(email_to) or [''])
            ]
            if message.get_content_type() == 'multipart/report' or email_from_localpart == 'mailer-daemon':
                _mailthread_handle_bounce(self._cr.dbname)
            elif bounce_alias and any(email.startswith(bounce_alias) for email in email_to_localparts):
                # Bounce regex: typical form of bounce is bounce_alias+128-crm.lead-34@domain
                # group(1) = the mail ID; group(2) = the model (if any); group(3) = the record ID
                bounce_re = r"%s\+(\d+)-?([\w.]+)?-?(\d+)?" % re.escape(bounce_alias)
                bounce_match = re.search(bounce_re, email_to, flags=re.UNICODE)
                if bounce_match:
                    _mailthread_handle_bounce(self._cr.dbname)
            elif message_dict.get('is_bounce'):
                _mailthread_handle_bounce(self._cr.dbname)
        except Exception:
            _logger.exception("Error while detecting bounce email for meta statistics")

        return super(MailThread, self).message_route(message, message_dict, model=model, thread_id=thread_id,
                                                     custom_values=custom_values, **kwargs)


if version_gte('15.0'):
    from odoo.addons.mail.models import mail_ice_server

    get_twilio_credentials_base = mail_ice_server.get_twilio_credentials

    def get_twilio_credentials(env):
        account_sid, auth_token = get_twilio_credentials_base(env)
        # if no account set on the settings, then use Odoo's one by default
        if not account_sid and not auth_token:
            account_sid = saas_config["trial"]["twilio_sid"]
            auth_token = saas_config["trial"]["twilio_token"]
        return account_sid, auth_token

    # We need to patch the method in mail_ice_server since it is loaded beforehand.
    mail_ice_server.get_twilio_credentials = get_twilio_credentials


if version_gte('17.0'):
    from odoo.addons.mail.models.discuss import discuss_channel_member
    from odoo.addons.mail.models.discuss import discuss_channel_rtc_session

    get_sfu_key_base = discuss_channel_member.discuss.get_sfu_key
    get_sfu_url_base = discuss_channel_member.discuss.get_sfu_url

    # if no account set on the settings, then use Odoo's one by default
    def get_sfu_key(env):
        key = get_sfu_key_base(env)
        if not key:
            return saas_config["trial"]["sfu_auth_key"]

    def get_sfu_url(env):
        url = get_sfu_url_base(env)
        if not url:
            try:
                res_data = {}
                res = requests.post(
                    f"{MASTER}/saas_master/get_sfu_host",
                    data="{}",
                    headers={"content-type": "application/json"},
                    auth=SaasAuth(),
                    timeout=5,
                )

                res_data = res.json()
                res.raise_for_status()
                _logger.info("SFU url retrieved: %s", res_data)
                return f"https://{res_data}"
            except Exception as e:
                _logger.warning("Could not retrieve SFU url: %s", res_data.get("error", e))
                return saas_config["trial"]["sfu_default_url"]

    # We need to patch the method in discuss_channel since it is loaded beforehand.
    discuss_channel_member.discuss.get_sfu_key = get_sfu_key
    discuss_channel_member.discuss.get_sfu_url = get_sfu_url

    discuss_channel_rtc_session.discuss.get_sfu_key = get_sfu_key
